# TODO to be remove as already implemented in alphaDIA.
from typing import Tuple, Union

import numpy as np
import pandas as pd
import tqdm
from alphatims.bruker import TimsTOF

from alpharaw.ms_data_base import MSData_Base, ms_reader_provider
from alpharaw.wrappers.alphatims_wrapper import AlphaTimsWrapper

from .psm_match import PepSpecMatch

alphatims_hdf_types = [
    "alphatims",
    "alphatims_hdf",
    "tims.hdf",
]


def load_ms_data_tims(
    ms_file: Union[str, MSData_Base, TimsTOF],
    ms_file_type: str = "alpharaw_hdf",
    dda: bool = False,
    spectra_sorted_by_rt: bool = True,
) -> Tuple[MSData_Base, TimsTOF]:
    """Load ms data as TimsTOF object

    Parameters
    ----------
    ms_file : str
        ms2 file path

    ms_file_type : str, optional
        ms2 file type, could be
        ["alpharaw_hdf","raw.hdf","thermo","sciex","alphapept_hdf","mgf"].
        Default to 'alphatims_hdf'

    dda : bool, optional
        if it is DDA data, by default False

    spectra_sorted_by_rt : bool, optional
        If spectra are already sorted by RT.
        Defaults to True

    Returns
    -------
    tuple
        MSData_Base: alpharaw MS Data (Reader) object
        TimsTOF: AlphaTims object
    """
    if isinstance(ms_file, TimsTOF):
        return None, ms_file
    elif ms_file_type.lower() in alphatims_hdf_types:
        return None, TimsTOF(ms_file)
    else:
        if isinstance(ms_file, MSData_Base):
            raw_data = ms_file
        else:
            raw_data = ms_reader_provider.get_reader(ms_file_type)
            raw_data.import_raw(ms_file)

            if not spectra_sorted_by_rt:
                # RT may not be sorted in AP HDF for timsTOF after preprocessing
                raw_data._sort_rt()

        tims_data = AlphaTimsWrapper(raw_data, dda=dda)
        return raw_data, tims_data


class PepSpecMatch_AlphaTims(PepSpecMatch):
    """
    Inherited from :class:`alpharaw.match.psm_match.PepSpecMatch`, but
    this can be used for DIA PSM matching by selecting
    MS2 spectra with RT (and IM) values.
    """

    #: RT win to get a MS2 spectrum by slicing
    rt_sec_tol_to_slice_ms2 = 3.0

    #: IM win to get a MS2 spectrum by slicing
    im_tol_to_slice_ms2 = 0.03

    #: find closest MS2 for the given RT when slicing
    find_k_nearest_ms2_by_rt = True
    k_rt_nearest = 1

    # : find closest MS2 for the given RT when slicing
    find_k_nearest_ms2_by_im = False
    k_im_nearest = 11

    def get_peak_df(
        self,
        precursor_mz: float,
        rt_sec: float,
        im: float = 0.0,
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        precursor_mz : float
            Precursor m/z value
        rt_sec : float
            RT value in seconds
        im : float, optional
            Ion mobility, by default 0.0

        Returns
        -------
        pd.DataFrame
            peak_df in alphatims DF format
        """
        rt_slice = slice(
            rt_sec - self.rt_sec_tol_to_slice_ms2,
            rt_sec + self.rt_sec_tol_to_slice_ms2,
        )

        if im == 0 or self.tims_data.scan_max_index == 1:
            im_slice = slice(None)
        elif self.find_k_nearest_ms2_by_im and self.tims_data.scan_max_index > 1:
            # AlphaTims without AlphaRaw for .d files
            im_slice = self.tims_data.scan_max_index - np.searchsorted(
                self.tims_data.mobility_values[::-1], im
            )
        else:
            im_slice = slice(
                im - self.im_tol_to_slice_ms2, im + self.im_tol_to_slice_ms2
            )

        spec_df = self.tims_data[rt_slice, im_slice, precursor_mz:precursor_mz]

        def find_k_nearest(array, val, k=3):
            nearest = np.argmin(np.abs(array - val))
            if nearest <= k // 2:
                return slice(k)
            elif nearest >= len(array) - k // 2 - 1:
                return slice(-k, None)
            else:
                return slice(nearest - k // 2, nearest + k // 2 + 1)

        if (
            self.find_k_nearest_ms2_by_im
            and im > 0
            and self.tims_data.scan_max_index > 1
        ):
            # RAW from AlphaRaw, mobility===0 in AlphaTims wrapper obj
            scan_idxes = np.sort(spec_df.scan_indices.unique())
            if len(scan_idxes) > 1:  # im from psm
                scan_idxes = scan_idxes[
                    find_k_nearest(
                        self.raw_data.spectrum_df.mobility.values[scan_idxes],
                        im,
                        self.k_im_nearest,
                    )
                ]
                spec_df = spec_df[spec_df.scan_indices.isin(scan_idxes)]

        if self.find_k_nearest_ms2_by_rt:
            rt_values = np.sort(spec_df.rt_values.unique())
            if len(rt_values) > 1:
                closest_rts = rt_values[
                    find_k_nearest(rt_values, rt_sec, self.k_rt_nearest)
                ]
                spec_df = spec_df[spec_df.rt_values.isin(closest_rts)]

        return spec_df

    def get_peaks(
        self,
        precursor_mz: float,
        rt_sec: float,
        im: float = 0.0,
    ) -> tuple:
        """
        Parameters
        ----------
        precursor_mz : float
            Precursor m/z value
        rt_sec : float
            RT value in seconds
        im : float, optional
            Ion mobility, by default 0.0

        Returns
        -------
        tuple
            np.ndarray: peak m/z values
            np.ndarray: peak intensity values
        """
        spec_df = self.get_peak_df(precursor_mz, rt_sec, im)
        spec_df = spec_df.sort_values("mz_values").reset_index(drop=True)
        return (spec_df.mz_values.values, spec_df.intensity_values.values)

    def load_ms_data(
        self,
        ms_file,
        ms_file_type,
        dda=False,
        spectra_sorted_by_rt=True,
    ):
        self.raw_data, self.tims_data = load_ms_data_tims(
            ms_file, ms_file_type, dda, spectra_sorted_by_rt
        )

    def match_ms2_one_raw(
        self, psm_df_one_raw: pd.DataFrame, verbose: bool = False
    ) -> tuple:
        """
        Matching psm_df_one_raw against
        self.tims_data and self.raw_data
        after `self.load_ms_data()`

        Parameters
        ----------
        psm_df_one_raw : pd.DataFrame
            psm dataframe
            that contains only one raw file

        Returns
        -------
        tuple:
            pd.DataFrame: psm dataframe with fragment index information.

            pd.DataFrame: fragment mz dataframe.

            pd.DataFrame: matched intensity dataframe.

            pd.DataFrame: matched mass error dataframe.
            np.inf if a fragment is not matched.

        """
        self.psm_df = psm_df_one_raw

        psm_df_one_raw = self._add_missing_columns_to_psm_df(psm_df_one_raw)

        (
            fragment_mz_df,
            matched_intensity_df,
            matched_mz_err_df,
        ) = self._prepare_matching_dfs()

        if (
            "mobility" in psm_df_one_raw.columns
            and "mobility" in self.raw_data.spectrum_df.columns
        ):
            query_columns = [
                "frag_start_idx",
                "frag_stop_idx",
                "precursor_mz",
                "rt",
                "mobility",
            ]
        else:
            query_columns = [
                "frag_start_idx",
                "frag_stop_idx",
                "precursor_mz",
                "rt",
            ]

        psm_iters = psm_df_one_raw[query_columns].values
        if verbose:
            psm_iters = tqdm.tqdm(psm_iters)

        for items in psm_iters:
            frag_start_idx = int(items[0])
            frag_stop_idx = int(items[1])

            spec_mzs, spec_intens = self.get_peaks(
                *items[2:],
            )
            self._match_one_psm(
                spec_mzs,
                spec_intens,
                fragment_mz_df,
                matched_intensity_df,
                matched_mz_err_df,
                frag_start_idx,
                frag_stop_idx,
            )
        return (psm_df_one_raw, fragment_mz_df, matched_intensity_df, matched_mz_err_df)

    def match_ms2_multi_raw(
        self,
        psm_df: pd.DataFrame,
        ms_files: Union[dict, list],
        ms_file_type: str = "alphatims",
        dda: bool = False,
    ):
        """
        Matching PSM dataframe against the ms2 files in ms_files
        This method will store matched values as attributes:
        - self.psm_df
        - self.fragment_mz_df
        - self.matched_intensity_df
        - self.matched_mz_err_df

        Parameters
        ----------
        psm_df : pd.DataFrame
            PSM dataframe

        ms_files : dict | list
            if dict: {raw_name: ms2 path}
            if list: [ms2 path1, ms2 path2]

        ms_file_type : str, optional
            One of ["alphatims_hdf","alpharaw_hdf","thermo","sciex","alphapept_hdf","mgf"]
            Defaults to 'alphapept'.

        Returns
        -------
        tuple:
            pd.DataFrame: psm dataframe with fragment index information.

            pd.DataFrame: fragment mz dataframe.

            pd.DataFrame: matched intensity dataframe.

            pd.DataFrame: matched mass error dataframe.
            np.inf if a fragment is not matched.

        """
        raise NotImplementedError(
            "Not necessary for matching multiple raw files using AlphaTims, "
            "loop through `match_ms2_one_raw()`"
        )



#!python

import warnings

warnings.filterwarnings("ignore")


def register_all_readers():
    from .legacy_msdata.mgf import register_readers as register_mgf_readers
    from .mzml import register_readers as register_mzml_readers

    register_mzml_readers()
    register_mgf_readers()

    try:
        from .sciex import register_readers as register_wiff_readers
        from .thermo import register_readers as register_raw_readers

        register_wiff_readers()
        register_raw_readers()
    except (RuntimeError, ImportError):
        print("[WARN] pythonnet is not installed")


__version__ = "0.4.7"



from typing import Tuple

import pandas as pd
from alphabase.peptide.fragment import (
    compress_fragment_indices,
    remove_unused_fragments,
)


def remove_unused_peaks(
    spectrum_df: pd.DataFrame,
    peak_df: pd.DataFrame,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Removes unused peaks of removed spectra,
    reannotates the peak_start_idx and peak_stop_idx

    Parameters
    ----------
    spectrum_df : pd.DataFrame
        Spectrum dataframe which contains peak_start_idx and peak_stop_idx columns

    peak_df : pd.DataFrame
        The peak dataframe which should be compressed by removing unused peaks.

    Returns
    -------
    pd.DataFrame, pd.DataFrame
        The reindexed spectrum DataFrame and the sliced peak DataFrame
    """

    spectrum_df, (peak_df,) = remove_unused_fragments(
        spectrum_df,
        (peak_df,),
        frag_start_col="peak_start_idx",
        frag_stop_col="peak_stop_idx",
    )
    return spectrum_df, peak_df

    spectrum_df = spectrum_df.sort_values(["peak_start_idx"], ascending=True)
    frag_idx = spectrum_df[["peak_start_idx", "peak_stop_idx"]].values

    new_frag_idx, fragment_pointer = compress_fragment_indices(frag_idx)

    spectrum_df[["peak_start_idx", "peak_stop_idx"]] = new_frag_idx
    spectrum_df = spectrum_df.sort_index()

    return (spectrum_df, peak_df.iloc[fragment_pointer].copy().reset_index(drop=True))



import logging

import numpy as np
import pandas as pd

from alpharaw.feature.chem import averagine_aa, isotopes, maximum_offset
from alpharaw.feature.hills import (
    extract_hills,
    filter_hills,
    get_hill_data,
    remove_duplicate_hills,
    split_hills,
)
from alpharaw.feature.isotope_pattern import (
    feature_finder_report,
    get_isotope_patterns,
    get_pre_isotope_patterns,
    get_stats,
)


def find(
    spectrum_df,
    peak_df,
    centroid_tol=8,
    max_gap=2,
    hill_length_min=3,
    hill_split_level=1.3,
    iso_split_level=1.3,
    hill_smoothing=1,
    hill_check_large=40,
    iso_charge_min=1,
    iso_charge_max=6,
    iso_n_seeds=100,
    hill_nboot_max=300,
    hill_nboot=150,
    iso_mass_range=5,
    iso_corr_min=0.6,
) -> pd.DataFrame:
    logging.info("FF started")
    query_data = df_to_ap_query_data(spectrum_df, peak_df)

    int_data = np.array(query_data["int_list_ms1"])

    logging.info(
        f"Hill extraction with centroid_tol {centroid_tol} and max_gap {max_gap}"
    )

    hill_ptrs, hill_data, path_node_cnt, score_median, score_std = extract_hills(
        query_data, max_gap, centroid_tol
    )
    logging.info(
        f"Number of hills {len(hill_ptrs):,}, len = {np.mean(path_node_cnt):.2f}"
    )

    logging.info(
        f"Repeating hill extraction with centroid_tol {score_median+score_std*3:.2f}"
    )

    hill_ptrs, hill_data, path_node_cnt, score_median, score_std = extract_hills(
        query_data, max_gap, score_median + score_std * 3
    )
    logging.info(
        f"Number of hills {len(hill_ptrs):,}, len = {np.mean(path_node_cnt):.2f}"
    )

    hill_ptrs, hill_data = remove_duplicate_hills(hill_ptrs, hill_data, path_node_cnt)
    logging.info(f"After duplicate removal of hills {len(hill_ptrs):,}")

    hill_ptrs = split_hills(
        hill_ptrs,
        hill_data,
        int_data,
        hill_split_level=hill_split_level,
        window=hill_smoothing,
    )  # hill lenght is inthere already
    logging.info(f"After split hill_ptrs {len(hill_ptrs):,}")

    hill_data, hill_ptrs = filter_hills(
        hill_data,
        hill_ptrs,
        int_data,
        hill_check_large=hill_check_large,
        window=hill_smoothing,
    )

    logging.info(f"After filter hill_ptrs {len(hill_ptrs):,}")

    stats, sortindex_, idxs_upper, scan_idx, hill_data, hill_ptrs = get_hill_data(
        query_data,
        hill_ptrs,
        hill_data,
        hill_nboot_max=hill_nboot_max,
        hill_nboot=hill_nboot,
    )
    logging.info("Extracting hill stats complete")

    pre_isotope_patterns = get_pre_isotope_patterns(
        stats,
        idxs_upper,
        sortindex_,
        hill_ptrs,
        hill_data,
        int_data,
        scan_idx,
        maximum_offset,
        iso_charge_min=iso_charge_min,
        iso_charge_max=iso_charge_max,
        iso_mass_range=iso_mass_range,
        cc_cutoff=iso_corr_min,
    )
    logging.info(f"Found {len(pre_isotope_patterns):,} pre isotope patterns.")

    isotope_patterns, iso_idx, isotope_charges = get_isotope_patterns(
        pre_isotope_patterns,
        hill_ptrs,
        hill_data,
        int_data,
        scan_idx,
        stats,
        sortindex_,
        averagine_aa,
        isotopes,
        iso_charge_min=iso_charge_min,
        iso_charge_max=iso_charge_max,
        iso_mass_range=iso_mass_range,
        iso_n_seeds=iso_n_seeds,
        cc_cutoff=iso_corr_min,
        iso_split_level=iso_split_level,
        callback=None,
    )
    logging.info(f"Extracted {len(isotope_charges):,} isotope patterns.")

    feature_df, lookup_idx = feature_finder_report(
        query_data,
        isotope_patterns,
        isotope_charges,
        iso_idx,
        stats,
        sortindex_,
        hill_ptrs,
        hill_data,
    )

    # lookup_idx_df = pd.DataFrame(lookup_idx, columns = ['isotope_pattern', 'isotope_pattern_hill'])

    feature_isotopes_df = get_stats(isotope_patterns, iso_idx, stats)

    feature_df["last_isotope_mz"] = (
        feature_isotopes_df.groupby("feature_id").tail(1).mz_average.values
    )

    logging.info("FF finished.")

    return feature_df


def df_to_ap_query_data(spec_df, peak_df):
    query_data = {}

    query_data["scan_list_ms1"] = spec_df.spec_idx.values + 1
    query_data["rt_list_ms1"] = spec_df.rt.values
    query_data["mass_list_ms1"] = peak_df.mz.values
    query_data["int_list_ms1"] = peak_df.intensity.values
    indices = np.zeros(len(spec_df) + 1, dtype=np.int64)
    indices[1:] = spec_df.peak_stop_idx.values
    query_data["indices_ms1"] = indices
    return query_data



import typing

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from alphatims.bruker import TimsTOF


def plot_scatter(
    x_values: np.ndarray,
    y_values: np.ndarray,
    color: str,
    marker_size: int,
    hovertext: str,
    hovertemplate: str,
    name: str = "",
) -> go.Scatter:
    """
    plotly scatter based on x and y points.

    Parameters
    ----------
    x_values : np.ndarray
        x values for :class:`go.Scatter`
    y_values : np.ndarray
        y values for :class:`go.Scatter`
    color : str
        color name
    marker_size : int
        marker size of the point (x,y)
    hovertext : str
        `hovertext` parameter of :class:`go.Scatter`
    hovertemplate : str
        `hovertemplate` parameter of :class:`go.Scatter`
    name : str, optional
        `name` parameter of :class:`go.Scatter`, by default ''

    Returns
    -------
    go.Scatter
        go.Scatter of plotly
    """
    return go.Scatter(
        x=x_values,
        y=y_values,
        hovertext=hovertext,
        hovertemplate=hovertemplate,
        mode="markers",
        marker=dict(color=color, size=marker_size),
        name=name,
        showlegend=False,
    )


alphatims_labels = {
    "rt": "rt_values",
    "im": "mobility_values",
}


def plot_line_tims(
    tims_data: TimsTOF,
    tims_raw_indices: np.ndarray,
    tims_view_indices: np.array,
    name: str,
    legend_group: str,
    marker_color: str,
    view_dim: typing.Literal["rt", "im"] = "rt",  # or 'im'
    intensity_scale: float = 1.0,
    rt_unit: str = "minute",
) -> go.Figure:
    """
    Plot an XIC line on alphatims `TimsTOF` data

    Parameters
    ----------
    tims_data : TimsTOF
        The alphatims `TimsTOF` object
    tims_raw_indices : np.ndarray
        Raw indices on `TimsTOF` object
    tims_view_indices : np.array
        View indices on `TimsTOF` object
    name : str
        Display name
    legend_group : str
        Lines will be grouped by `legend_group`
    marker_color : str
        Color of the scatter (x, y)
    view_dim : "rt", "im", optional
        View dimension, "rt" or "im", by default "rt"
    rt_unit : str, optional
        RT unit, by default "minute"

    Returns
    -------
    go.Figure
        Line plot
    """
    x_dimension = alphatims_labels[view_dim]

    intensities = tims_data.bin_intensities(tims_raw_indices, [x_dimension])
    if view_dim == "rt":
        x_ticks = tims_data.rt_values[tims_view_indices]
        intensities = intensities[tims_view_indices]
        if rt_unit == "minute":
            x_ticks /= 60.0
    else:
        x_ticks = tims_data.mobility_values[tims_view_indices]
        intensities = intensities[tims_view_indices]

    return plot_line(
        x_ticks,
        intensities * intensity_scale,
        name=name,
        marker_color=marker_color,
        legend_group=legend_group,
        x_text=view_dim.upper(),
    )


def plot_line_tims_fast(
    tims_data: TimsTOF,
    tims_raw_indices: np.ndarray,
    tims_view_indices: np.array,
    name: str,
    legend_group: str,
    marker_color: str,
    view_dim: typing.Literal["rt", "im"] = "rt",
    intensity_scale: float = 1.0,
    rt_unit: str = "minute",
    add_peak_area=True,
) -> go.Figure:
    """
    Plot an XIC line on alphatims `TimsTOF` data

    Parameters
    ----------
    tims_data : TimsTOF
        The alphatims `TimsTOF` object
    tims_raw_indices : np.ndarray
        Raw indices on `TimsTOF` object
    tims_view_indices : np.array
        View indices on `TimsTOF` object
    name : str
        Display name
    legend_group : str
        Lines will be grouped by `legend_group`
    marker_color : str
        Color of the scatter (x, y)
    view_dim : "rt", "im", optional
        View dimension, "rt" or "im", by default "rt"
    intensity_scale : float, optional
        Intensity scale of mirror plot, by default 1.0
    rt_unit : str, optional
        RT unit, by default "minute"
    add_peak_area : bool, optional
        If add peak area in the hover text, by default True

    Returns
    -------
    go.Figure
        _description_
    """
    x_dimension = alphatims_labels[view_dim]

    intensities = tims_data.bin_intensities(tims_raw_indices, [x_dimension])
    if view_dim == "rt":
        x_ticks = tims_data.rt_values[tims_view_indices]
        intensities = intensities[tims_view_indices]
        if rt_unit == "minute":
            x_ticks /= 60.0
    else:
        x_ticks = tims_data.mobility_values[tims_view_indices]
        intensities = intensities[tims_view_indices]

    if add_peak_area:
        peak_area = abs(np.trapz(y=intensities, x=x_ticks))

    return plot_line(
        x_ticks,
        intensities * intensity_scale,
        name=name,
        marker_color=marker_color,
        legend_group=legend_group,
        x_text=view_dim.upper(),
        other_info=f"Peak area: {peak_area:.2e}" if add_peak_area else "",
    )


def plot_line(
    x_values: np.ndarray,
    y_values: np.ndarray,
    name: str,
    marker_color: str,
    legend_group: str,
    x_text: str = "RT",
    other_info: str = "",
    hovertemplate: str = "%{text} <br>Intensity: %{y}",
) -> go.Scatter:
    """
    Plot a line for given x and y points, this is usually for XIC plots.

    Parameters
    ----------
    x_values : np.ndarray
        x values for :class:`go.Scatter`
    y_values : np.ndarray
        y values for :class:`go.Scatter`
    name : str
        Legend name for the plotted line
    marker_color : str
        Marker color of the point (x,y)
    legend_group : str
        Different lines can be grouped by the same `legend_group`
    x_text : str, optional
        Hover text of the x axis, by default "RT"
    other_info: str, optional
        Other hover information, by default ""
    hovertemplate : str, optional
        `hovertemplate` parameter of :class:`go.Scatter`,
        by default '%{text} <br><b>Intensity:</b> %{y}'

    Returns
    -------
    go.Scatter
        The plotted line
    """
    return go.Scatter(
        x=x_values,
        y=y_values,
        mode="lines",
        name=name,
        marker={"color": marker_color},
        legendgroup=legend_group,
        text=[
            f"{x_text}: {_x:.3f}" + (f"<br>{other_info}" if other_info else "")
            for _x in x_values
        ],
        hovertemplate=hovertemplate,
    )


def plot_line_df(
    tims_sliced_df: pd.DataFrame,
    view_indices_df: pd.DataFrame,
    label: str,
    legend_group: str,
    marker_color: str,
    view_dim: str = "rt",  # or 'im'
) -> go.Figure:
    """
    TODO deprecated
    """
    if view_dim == "rt":
        tims_sliced_df = tims_sliced_df.groupby("frame_indices", as_index=False).agg(
            {
                "rt_values": "mean",
                "intensity_values": "sum",
            }
        )
        tims_sliced_df["rt_values"] /= 60
        tims_sliced_df.sort_values("rt_values", inplace=True)
        tims_sliced_df = view_indices_df.merge(
            tims_sliced_df, on=["frame_indices", "rt_values"], how="left"
        )
        tims_sliced_df.loc[
            tims_sliced_df.intensity_values.isna(), "intensity_values"
        ] = 0
        x_ticks = tims_sliced_df.rt_values.values
    else:
        tims_sliced_df = tims_sliced_df.groupby("scan_indices", as_index=False).agg(
            {
                "mobility_values": "mean",
                "intensity_values": "sum",
            }
        )
        tims_sliced_df.sort_values("mobility_values", inplace=True)
        tims_sliced_df = view_indices_df.merge(
            tims_sliced_df, on=["scan_indices", "mobility_values"], how="left"
        )
        tims_sliced_df.loc[
            tims_sliced_df.intensity_values.isna(), "intensity_values"
        ] = 0
        x_ticks = tims_sliced_df.mobility_values.values

    trace = go.Scatter(
        x=x_ticks,
        y=tims_sliced_df.intensity_values.values,
        mode="lines",
        text=[f"RT: {_x*60:.3f}s" for _x in x_ticks]
        if view_dim == "rt"
        else [f"IM: {_x:.3f}" for _x in x_ticks],
        hovertemplate="%{text}<br>Intensity: %{y}",
        name=label,
        marker={"color": marker_color},
        legendgroup=legend_group,
    )
    return trace



from typing import Tuple

import numpy as np
from numba import njit


@njit
def naive_centroid(
    peak_mzs: np.ndarray,
    peak_intensities: np.ndarray,
    centroiding_ppm: float = 20.0,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    A naive centroiding algorithm.

    Parameters
    ----------
    peak_mzs : np.ndarray
        peak m/z values to centroid.
    peak_intensities : np.ndarray
        peak intensities to centroid.
    centroiding_ppm : float, optional
        The centroiding ppm, by default 20.0

    Returns
    -------
    Tuple
        ndarray: peak m/z array
        ndarray: peak intensity array
    """
    mz_list = []
    inten_list = []
    start, stop = 0, 1
    centroiding_peak_tols = 2 * peak_mzs * centroiding_ppm * 1e-6
    while start < len(peak_mzs):
        stop = _find_sister_peaks(peak_mzs, centroiding_peak_tols, start)
        mz_list.append(
            np.average(peak_mzs[start:stop], weights=peak_intensities[start:stop])
        )
        inten_list.append(np.sum(peak_intensities[start:stop]))
        start = stop
    return (
        np.array(mz_list, dtype=peak_mzs.dtype),
        np.array(inten_list, dtype=peak_intensities.dtype),
    )


@njit
def _find_sister_peaks(
    peak_mzs: np.ndarray, centroiding_peak_tols: np.ndarray, start: int
):
    """
    Find sister peak stop idx for the given start idx.
    Sister peaks refers to peaks from the same ion in profile mode.
    Internal function.
    """
    stop = start + 1
    for i in range(start + 1, len(peak_mzs)):
        if peak_mzs[i] - peak_mzs[start] <= centroiding_peak_tols[start]:
            stop = i + 1
        else:
            stop = i
            break
    return stop



import warnings

import numpy as np
import pandas as pd
from sklearn.neighbors import KNeighborsRegressor

warnings.warn(
    "This module will be removed in the future as "
    "mass calibration has already been implemented in alphaDIA.",
    category=DeprecationWarning,
)


def get_fragment_median(start_end_idxes: tuple, frag_df: pd.DataFrame):
    start_idx, end_idx = start_end_idxes
    ret = np.nanmedian(frag_df.values[start_idx:end_idx])
    if np.isnan(ret):
        return 0.0
    else:
        return ret


def calibrate_one(start_end_shift, frag_df):
    start_idx, end_idx, mass_shift = start_end_shift
    frag_df.values[int(start_idx) : int(end_idx)] -= mass_shift


class MassCalibratorForRT_KNN:
    """Using KNN to calibrate measured m/z across RT."""

    def __init__(self, n_neighbors=5):
        self._n_neighbors = n_neighbors
        self.model = KNeighborsRegressor(n_neighbors)

    def fit(self, psm_df: pd.DataFrame, mass_error_df: pd.DataFrame):
        mass_error_df = mass_error_df.replace(np.inf, np.nan)
        mean_merrs = (
            psm_df[["frag_start_idx", "frag_stop_idx"]]
            .apply(get_fragment_median, axis=1, frag_df=mass_error_df)
            .values
        )
        self.model.fit(psm_df.rt.values.reshape((-1, 1)), mean_merrs.reshape(-1, 1))

    def calibrate(
        self, psm_df: pd.DataFrame, mass_error_df: pd.DataFrame
    ) -> pd.DataFrame:
        psm_df["frag_mass_shift"] = self.model.predict(
            psm_df.rt.values.reshape((-1, 1))
        ).reshape(-1)
        psm_df[["frag_start_idx", "frag_stop_idx", "frag_mass_shift"]].apply(
            calibrate_one, axis=1, frag_df=mass_error_df
        )
        return mass_error_df
